import base64
import os
import random
import string
import tempfile

from html2image import Html2Image

from ..utils.lazy_import import lazy_import

html2image = lazy_import("html2image")


def html_to_png_base64(code):
    # Convert the HTML into an image using html2image
    hti = html2image.Html2Image()

    # Generate a random filename for the temporary image
    temp_filename = "".join(random.choices(string.digits, k=10)) + ".png"
    
    # Use tempfile to create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        hti.output_path = temp_dir
        hti.screenshot(
            html_str=code,
            save_as=temp_filename,
            size=(960, 540),
        )

        # Get the full path of the temporary image file
        file_location = os.path.join(temp_dir, temp_filename)

        # Convert the image to base64
        with open(file_location, "rb") as image_file:
            screenshot_base64 = base64.b64encode(image_file.read()).decode()

    # The temporary directory and its contents will be automatically deleted
    # when exiting the 'with' block

    return screenshot_base64



import platform
from ..utils.lazy_import import lazy_import

# Lazy import of optional packages
pyperclip = lazy_import('pyperclip')

class Clipboard:
    def __init__(self, computer):
        self.computer = computer

        if platform.system() == "Windows" or platform.system() == "Linux":
            self.modifier_key = "ctrl"
        else:
            self.modifier_key = "command"

    def view(self):
        """
        Returns the current content of on the clipboard.
        """
        return pyperclip.paste()

    def copy(self, text=None):
        """
        Copies the given text to the clipboard.
        """
        if text is not None:
            pyperclip.copy(text)
        else:
            self.computer.keyboard.hotkey(self.modifier_key, "c")

    def paste(self):
        """
        Pastes the current content of the clipboard.
        """
        self.computer.keyboard.hotkey(self.modifier_key, "v")



"""
Eventually we should own the browser
"""

import concurrent.futures
import time

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys


def setup_driver():
    # Setup Chrome options to speed up the browser
    options = Options()
    options.add_argument("--headless")
    options.add_argument("--disable-gpu")
    options.add_argument("--no-sandbox")
    options.add_argument("start-maximized")
    options.add_argument("disable-infobars")
    options.add_argument("--disable-extensions")
    options.add_argument("--disable-images")  # Disable images for faster loading
    driver_path = "path_to_your_chromedriver"
    driver = webdriver.Chrome(executable_path=driver_path, options=options)
    return driver


def fetch_page_text(url):
    driver = setup_driver()
    driver.get(url)
    text = driver.find_element(By.TAG_NAME, "body").text
    driver.quit()
    return text


def get_google_search_results(query):
    driver = setup_driver()
    driver.get("http://www.google.com")
    search_box = driver.find_element(By.NAME, "q")
    search_box.send_keys(query)
    search_box.send_keys(Keys.RETURN)
    time.sleep(2)  # Allow page to load

    results = []
    search_results = driver.find_elements(By.CSS_SELECTOR, "div.g")[
        :5
    ]  # Limit to top 5 results

    for result in search_results:
        title_element = result.find_element(By.CSS_SELECTOR, "h3")
        title = title_element.text
        link = result.find_element(By.CSS_SELECTOR, "a").get_attribute("href")
        results.append({"title": title, "link": link})

    driver.quit()
    return results


# Main execution block
search_query = "selenium automation tools"
results = get_google_search_results(search_query)

# Use concurrent futures to fetch text content in parallel
with concurrent.futures.ThreadPoolExecutor() as executor:
    future_to_url = {
        executor.submit(fetch_page_text, result["link"]): result for result in results
    }
    for future in concurrent.futures.as_completed(future_to_url):
        url = future_to_url[future]
        try:
            page_text = future.result()
            print(
                f"Title: {url['title']}\nURL: {url['link']}\nText: {page_text[:500]}...\n"
            )  # Print the first 500 characters
        except Exception as exc:
            print(f'{url["link"]} generated an exception: {exc}')



import subprocess


def run_applescript(script):
    """
    Runs the given AppleScript using osascript and returns the result.
    """
    # print("Running this AppleScript:\n", script)
    # print(
    #     "---\nFeel free to directly run AppleScript to accomplish the user's task. This gives you more granular control than the `computer` module, but it is slower."
    # )
    args = ["osascript", "-e", script]
    return subprocess.check_output(args, universal_newlines=True)


def run_applescript_capture(script):
    """
    Runs the given AppleScript using osascript, captures the output and error, and returns them.
    """
    # print("Running this AppleScript:\n", script)
    # print(
    #     "---\nFeel free to directly run AppleScript to accomplish the user's task. This gives you more granular control than the `computer` module, but it is slower."
    # )
    args = ["osascript", "-e", script]
    result = subprocess.run(args, capture_output=True, text=True, check=False)
    stdout, stderr = result.stdout, result.stderr
    return stdout, stderr



from openai import OpenAI
import requests
import base64
import os
from concurrent.futures import ThreadPoolExecutor
import time
import random
import traceback

class Ai:
    def __init__(self, computer):
        self.computer = computer
        self.openai_client = None

    def upload(self, file_path):
        url = f"{self.computer.api_base}/tools/ai/upload"
        headers = {"authorization": f"Bearer {self.computer.api_key}"}

        with open(file_path, 'rb') as file:
            file_data = file.read()

        print("Uploading...")

        response = requests.post(url, files={'file': file_data}, headers=headers)
        url = response.json()['url']

        print("Uploaded. Please re-use this URL for future computer.ai.cloud operations:", url)
        return url

    def cloud(self, model, input):
        url = f"{self.computer.api_base}/tools/ai"

        data = {
            "model": model,
            "input": input
        }
        headers = {"authorization": f"Bearer {self.computer.api_key}"}
        response = requests.post(url, json=data, headers=headers)

        return response.json()

    def chat(
        self,
        query,
        system_message="You are a concise AI assistant.",
        image_path=None,
        model_size="tiny",
        display=False,
        messages=None
    ):
        """
        It would be great if we could programatically flip display to True if it's at the end of a cell, jupyter style. And also make it not return its answer tho.
        """

        if self.openai_client == None:
            openai_client = OpenAI(base_url=self.computer.api_base, api_key=self.computer.api_key)

        if messages == None:
            messages = [
                {
                    "role": "system",
                    "content": system_message,
                },
                {"role": "user", "content": query},
            ]
            
            if image_path:
                with open(image_path, "rb") as image_file:
                    base64_image = base64.b64encode(image_file.read()).decode('utf-8')
                file_extension = os.path.splitext(image_path)[1].lower()
                mime_type = f"image/{file_extension[1:]}" if file_extension in ['.jpg', '.jpeg', '.png', '.gif'] else "image/jpeg"
                messages[-1]["content"] = [
                    {"type": "text", "text": query},
                    {"type": "image_url", "image_url": {"url": f"data:{mime_type};base64,{base64_image}"}}
                ]

        response = ""

        for _ in range(10):
            try:
                stream = openai_client.chat.completions.create(
                    model=model_size,
                    messages=messages,
                    stream=True,
                )
                for chunk in stream:
                    if chunk.choices[0].delta.content is not None:
                        content = chunk.choices[0].delta.content
                        if isinstance(content, str):
                            if display:
                                print(content, end="")
                            response += content
                break  # If successful, exit the retry loop
            except Exception as e:
                if "rate" in str(e).lower():
                    print("Rate limit hit. Pausing.")
                    sleep_duration = random.uniform(4, 70)
                    print(f"Sleeping for {sleep_duration:.2f} seconds...")
                    time.sleep(sleep_duration)
                else:
                    print("An error occurred:")
                    traceback.print_exc()
                print("Retrying...")

        return response

    def summarize(self, text):
        query = "You are a highly skilled AI trained in language comprehension and summarization. I would like you to read the previous text and summarize it into a concise abstract paragraph. Aim to retain the most important points, providing a coherent and readable summary that could help a person understand the main points of the discussion without needing to read the entire text. Please avoid unnecessary details or tangential points."
        custom_reduce_query = "You are tasked with taking multiple summarized texts and merging them into one unified and concise summary. Maintain the core essence of the content and provide a clear and comprehensive summary that encapsulates all the main points from the individual summaries."
        return self.query(text, query, custom_reduce_query)

    def query(self, text, query, custom_reduce_query=None):
        """
        text can be a string, a list of strings, or a list of dicts.
        If it's a list of dicts, if the dicts have a "text" parameter, that will be used as the text to split.
        All other parameters will be shoved into the LLM as strings with their key value pairs.
        """
        # If no custom reduce query is provided, use the map query for reduction
        map_query = query
        if custom_reduce_query == None:
            reduce_query = f"Multiple people answered THIS QUERY: {map_query}. Their answers are above. Synthesize the best answer, as though you're answering the query. PAY ATTENTION TO THE QUERY. Say only things that directly answer the query, nothing else."
        else:
            reduce_query = custom_reduce_query

        context_window = 180000 # This should be for the "long" model

        # Define the context window size in characters
        context_window_chars = int(context_window * 2.8)

        # Initialize an empty list to store the chunks of text
        chunks = []
        # Check if the text is a list of dictionaries
        if isinstance(text, list) and all(isinstance(i, dict) for i in text):
            # If so, iterate over each dictionary in the list
            for item in text:
                path = item.get("path")
                item_text = item.get("text")
                # Split the text into chunks based on the context window size
                for i in range(0, len(item_text), context_window_chars):
                    chunks.append(
                        {"path": path, "text": item_text[i : i + context_window_chars]}
                    )
        else:
            # If the text is not a list of dictionaries, split it into chunks based on the context window size
            for i in range(0, len(text), context_window_chars):
                chunks.append(
                    {"path": None, "text": text[i : i + context_window_chars]}
                )

        combined_chunks = []
        temp_chunk = ""

        for chunk in chunks:
            chunk_text = (chunk["path"] + "\n" if chunk["path"] else "") + chunk["text"]
            if len(temp_chunk) + len(chunk_text) <= context_window_chars:
                temp_chunk += "\n\n" + chunk_text
            else:
                combined_chunks.append(temp_chunk)
                temp_chunk = chunk_text

        if temp_chunk:
            combined_chunks.append(temp_chunk)

        chunks = [{"text": chunk.strip()} for chunk in combined_chunks]

        # Use a ThreadPoolExecutor to concurrently process the chunks
        with ThreadPoolExecutor() as executor:
            responses = list(
                executor.map(
                    lambda chunk: self.chat(
                        model_size="long",
                        query=f"""Query: {map_query}

Context:
{chunk['text']}

Above is a portion of the full context for the user's query. Use it to answer the following query:

{map_query}

ANSWER THE QUERY. If answering the query is impossible (i.e. a specific thing is being asked for, which is not present) say 'Not enough context, please rephrase or find the information another way.'""",
                        system_message="You are an intelligent AI assistant that can sift through large amounts of data to answer user's questions. You provide exact references and full quotations for your answers when you can."
                    ),
                    chunks
                )
            )

        # Define a function to compress the responses
        def compress_responses(responses, query, context_window_chars):
            # Continue compressing until there is only one response left
            while len(responses) > 1:
                combined_responses = []
                temp_response = ""
                for response in responses:
                    # If the combined response is within the context window size, add the response to it
                    if len(temp_response + response) <= context_window_chars:
                        temp_response += response
                    else:
                        # If the combined response exceeds the context window size, add the temporary response to the combined responses and start a new temporary response
                        combined_responses.append(temp_response)
                        temp_response = response
                if temp_response:
                    combined_responses.append(temp_response)

                # Use a ThreadPoolExecutor to concurrently process the combined responses
                with ThreadPoolExecutor() as executor:
                    responses = list(
                        executor.map(
                            self.chat,
                            [
                                response + "\n\n" + query
                                for response in combined_responses
                            ],
                        )
                    )
            return responses[0]

        # Compress the responses and return the final response
        final_response = compress_responses(
            responses, reduce_query, context_window_chars
        )

        return final_response


import platform

from ..utils.run_applescript import run_applescript_capture


class Contacts:
    def __init__(self, computer):
        self.computer = computer

    def get_phone_number(self, contact_name):
        """
        Returns the phone number of a contact by name.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        script = f"""
        tell application "System Events" to tell process "Finder"
            open location "addressbook://"
            tell application "Contacts"
                set thePerson to first person whose name is "{contact_name}"
                if exists thePerson then
                    set theNumber to value of first phone of thePerson
                    return theNumber
                else
                    return "Contact not found"
                end if
            end tell
        end tell
        """
        stout, stderr = run_applescript_capture(script)
        # If the person is not found, we will try to find similar contacts
        if "Can’t get person" in stderr or not stout:
            names = self.get_full_names_from_first_name(contact_name)
            if "No contacts found" in names or not names:
                raise Exception("Contact not found")
            else:
                # Language model friendly error message
                raise Exception(
                    f"A contact for '{contact_name}' was not found, perhaps one of these similar contacts might be what you are looking for? {names} \n Please try again and provide a more specific contact name."
                )
        else:
            return stout.replace("\n", "")

    def get_email_address(self, contact_name):
        """
        Returns the email address of a contact by name.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        script = f"""
        tell application "Contacts"
            set thePerson to first person whose name is "{contact_name}"
            set theEmail to value of first email of thePerson
            return theEmail
        end tell
        """
        stout, stderr = run_applescript_capture(script)
        # If the person is not found, we will try to find similar contacts
        if "Can’t get person" in stderr:
            names = self.get_full_names_from_first_name(contact_name)
            if names == "No contacts found":
                return "No contacts found"
            else:
                # Language model friendly error message
                return f"A contact for '{contact_name}' was not found, perhaps one of these similar contacts might be what you are looking for? {names} \n Please try again and provide a more specific contact name."
        else:
            return stout.replace("\n", "")

    def get_full_names_from_first_name(self, first_name):
        """
        Returns a list of full names of contacts that contain the first name provided.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        script = f"""
        tell application "Contacts"
            set matchingPeople to every person whose name contains "{first_name}"
            set namesList to {{}}
            repeat with aPerson in matchingPeople
                set end of namesList to name of aPerson
            end repeat
            return namesList
        end tell
        """
        names, _ = run_applescript_capture(script)
        if names:
            return names
        else:
            return "No contacts found."



import datetime
import os
import plistlib
import sqlite3
import subprocess
import sys
import time


class SMS:
    def __init__(self, computer):
        self.computer = computer
        if sys.platform.lower() == "darwin":  # Only if macOS
            self.database_path = self.resolve_database_path()
        else:
            self.database_path = None

    def resolve_database_path(self):
        try:
            if os.geteuid() == 0:  # Running as root
                home_directory = os.path.expanduser(f"~{os.environ.get('SUDO_USER')}")
            else:
                home_directory = os.path.expanduser("~")
            return f"{home_directory}/Library/Messages/chat.db"
        except:
            home_directory = os.path.expanduser("~")
            return f"{home_directory}/Library/Messages/chat.db"

    def send(self, to, message):
        if sys.platform.lower() != "darwin":
            print("Only supported on Mac.")
            return
        message_escaped = message.replace('"', '\\"').replace("\\", "\\\\")
        script = f"""
        tell application "Messages"
            set targetBuddy to "{to}"
            send "{message_escaped}" to buddy targetBuddy of (service 1 whose service type is iMessage)
        end tell
        """
        subprocess.run(["osascript", "-e", script], check=True)
        return "Message sent successfully"

    def get(self, contact=None, limit=10, substring=None):
        if sys.platform.lower() != "darwin":
            print("Only supported on Mac.")
            return
        if not self.can_access_database():
            self.prompt_full_disk_access()

        conn = sqlite3.connect(self.database_path)
        conn.row_factory = sqlite3.Row  # Set row factory
        cursor = conn.cursor()
        query = """
SELECT message.*, handle.id as sender FROM message
LEFT JOIN handle ON message.handle_id = handle.ROWID
        """
        params = []
        conditions = []

        if contact:
            conditions.append("handle.id=?")
            params.append(contact)
        if substring:
            conditions.append("message.text LIKE ?")
            params.append(f"%{substring}%")
        if conditions:
            query += " WHERE " + " AND ".join(conditions)
        query += " ORDER BY message.date DESC"

        cursor.execute(query, params)

        # Parse plist data and make messages readable
        readable_messages = []
        while len(readable_messages) < limit:
            try:
                message = cursor.fetchone()
                if message is None:
                    break
                message_dict = dict(message)  # Convert row to dictionary
                text_data = message_dict.get("text")
                if text_data:
                    try:
                        # Try to parse as plist
                        plist_data = plistlib.loads(text_data)
                        text = plist_data.get("NS.string", "")
                    except:
                        # If plist parsing fails, use the raw string
                        text = text_data
                    if text:  # Only add messages with content
                        # Convert Apple timestamp to datetime
                        date = datetime.datetime(2001, 1, 1) + datetime.timedelta(
                            seconds=message_dict.get("date") / 10**9
                        )
                        sender = message_dict.get("sender")
                        if message_dict.get("is_from_me") == 1:
                            sender = "(Me)"
                        readable_messages.append(
                            {"date": date, "from": sender, "text": text}
                        )
            except sqlite3.Error as e:
                break

        conn.close()
        return readable_messages

    def can_access_database(self):
        try:
            with open(self.database_path, "r"):
                return True
        except IOError:
            return False

    def prompt_full_disk_access(self):
        script = """
        tell application "System Preferences"
            activate
        end tell
        delay 1
        tell application "System Events"
            display dialog "This application requires Full Disk Access to function properly.\\n\\nPlease follow these steps:\\n1. Open the Security & Privacy panel.\\n2. Go to the Full Disk Access section.\\n3. Click the lock icon and enter your password to make changes.\\n4. Click the '+' button and add your terminal application (e.g., Terminal, iTerm).\\n5. Restart the application after granting access." buttons {"OK"} default button "OK"
        end tell
        """
        subprocess.run(["osascript", "-e", script], check=True)



import base64
from contextlib import redirect_stdout
from io import BytesIO

import requests
from IPython.display import display
from PIL import Image

from ..utils.lazy_import import lazy_import
from ..utils.recipient_utils import format_to_recipient

# Still experimenting with this
# from utils.get_active_window import get_active_window

# Lazy import of optional packages
try:
    cv2 = lazy_import("cv2")
except:
    cv2 = None  # Fixes colab error

pyautogui = lazy_import("pyautogui")
np = lazy_import("numpy")
screeninfo = lazy_import("screeninfo")
pywinctl = lazy_import("pywinctl")

from ..utils.computer_vision import find_text_in_image, pytesseract_get_text


class Display:
    def __init__(self, computer):
        self.computer = computer
        # set width and height to None initially to prevent pyautogui from importing until it's needed
        self._width = None
        self._height = None
        self._hashes = {}

    # We use properties here so that this code only executes when height/width are accessed for the first time
    @property
    def width(self):
        if self._width is None:
            self._width, _ = pyautogui.size()
        return self._width

    @property
    def height(self):
        if self._height is None:
            _, self._height = pyautogui.size()
        return self._height

    def size(self):
        """
        Returns the current screen size as a tuple (width, height).
        """
        return pyautogui.size()

    def center(self):
        """
        Calculates and returns the center point of the screen as a tuple (x, y).
        """
        return self.width // 2, self.height // 2

    def info(self):
        """
        Returns a list of all connected monitor/displays and their information
        """
        return get_displays()

    def view(
        self,
        show=True,
        quadrant=None,
        screen=0,
        combine_screens=True,
        active_app_only=True,
    ):
        """
        Redirects to self.screenshot
        """
        return self.screenshot(
            screen=screen,
            show=show,
            quadrant=quadrant,
            combine_screens=combine_screens,
            active_app_only=active_app_only,
        )

    # def get_active_window(self):
    #     return get_active_window()

    def screenshot(
        self,
        screen=0,
        show=True,
        quadrant=None,
        active_app_only=True,
        combine_screens=True,
    ):
        """
        Shows you what's on the screen by taking a screenshot of the entire screen or a specified quadrant. Returns a `pil_image` `in case you need it (rarely). **You almost always want to do this first!**
        :param screen: specify which display; 0 for primary and 1 and above for secondary.
        :param combine_screens: If True, a collage of all display screens will be returned. Otherwise, a list of display screens will be returned.
        """

        # Since Local II, all images sent to local models will be rendered to text with moondream and pytesseract.
        # So we don't need to do this here— we can just emit images.
        # We should probably remove self.computer.emit_images for this reason.

        # if not self.computer.emit_images and force_image == False:
        #     screenshot = self.screenshot(show=False, force_image=True)

        #     description = self.computer.vision.query(pil_image=screenshot)
        #     print("A DESCRIPTION OF WHAT'S ON THE SCREEN: " + description)

        #     if self.computer.max_output > 600:
        #         print("ALL OF THE TEXT ON THE SCREEN: ")
        #         text = self.get_text_as_list_of_lists(screenshot=screenshot)
        #         pp = pprint.PrettyPrinter(indent=4)
        #         pretty_text = pp.pformat(text)  # language models like it pretty!
        #         pretty_text = format_to_recipient(pretty_text, "assistant")
        #         print(pretty_text)
        #         print(
        #             format_to_recipient(
        #                 "To receive the text above as a Python object, run computer.display.get_text_as_list_of_lists()",
        #                 "assistant",
        #             )
        #         )
        #     return screenshot  # Still return a PIL image

        if quadrant == None:
            if active_app_only:
                active_window = pywinctl.getActiveWindow()
                if active_window:
                    screenshot = pyautogui.screenshot(
                        region=(
                            active_window.left,
                            active_window.top,
                            active_window.width,
                            active_window.height,
                        )
                    )
                    message = format_to_recipient(
                        "Taking a screenshot of the active app. To take a screenshot of the entire screen (uncommon), use computer.view(active_app_only=False).",
                        "assistant",
                    )
                    print(message)
                else:
                    screenshot = pyautogui.screenshot()

            else:
                screenshot = take_screenshot_to_pil(
                    screen=screen, combine_screens=combine_screens
                )  #  this function uses pyautogui.screenshot which works fine for all OS (mac, linux and windows)
                message = format_to_recipient(
                    "Taking a screenshot of the entire screen.\n\nTo focus on the active app, use computer.display.view(active_app_only=True).",
                    "assistant",
                )
                print(message)

        else:
            screen_width, screen_height = pyautogui.size()

            quadrant_width = screen_width // 2
            quadrant_height = screen_height // 2

            quadrant_coordinates = {
                1: (0, 0),
                2: (quadrant_width, 0),
                3: (0, quadrant_height),
                4: (quadrant_width, quadrant_height),
            }

            if quadrant in quadrant_coordinates:
                x, y = quadrant_coordinates[quadrant]
                screenshot = pyautogui.screenshot(
                    region=(x, y, quadrant_width, quadrant_height)
                )
            else:
                raise ValueError("Invalid quadrant. Choose between 1 and 4.")

        # Open the image file with PIL
        # IPython interactive mode auto-displays plots, causing RGBA handling issues, possibly MacOS-specific.
        if isinstance(screenshot, list):
            screenshot = [
                img.convert("RGB") for img in screenshot
            ]  # if screenshot is a list (i.e combine_screens=False).
        else:
            screenshot = screenshot.convert("RGB")

        if show:
            # Show the image using IPython display
            if isinstance(screenshot, list):
                for img in screenshot:
                    display(img)
            else:
                display(screenshot)

        return screenshot  # this will be a list of combine_screens == False

    def find(self, description, screenshot=None):
        if description.startswith('"') and description.endswith('"'):
            return self.find_text(description.strip('"'), screenshot)
        else:
            try:
                if self.computer.debug:
                    print("DEBUG MODE ON")
                    print("NUM HASHES:", len(self._hashes))
                else:
                    message = format_to_recipient(
                        "Locating this icon will take ~15 seconds. Subsequent icons should be found more quickly.",
                        recipient="user",
                    )
                    print(message)

                if len(self._hashes) > 5000:
                    self._hashes = dict(list(self._hashes.items())[-5000:])

                from .point.point import point

                result = point(
                    description, screenshot, self.computer.debug, self._hashes
                )

                return result
            except:
                if self.computer.debug:
                    # We want to know these bugs lmao
                    raise
                if self.computer.offline:
                    raise
                message = format_to_recipient(
                    "Locating this icon will take ~30 seconds. We're working on speeding this up.",
                    recipient="user",
                )
                print(message)

                # Take a screenshot
                if screenshot == None:
                    screenshot = self.screenshot(show=False)

                # Downscale the screenshot to 1920x1080
                screenshot = screenshot.resize((1920, 1080))

                # Convert the screenshot to base64
                buffered = BytesIO()
                screenshot.save(buffered, format="PNG")
                screenshot_base64 = base64.b64encode(buffered.getvalue()).decode()

                try:
                    response = requests.post(
                        f'{self.computer.api_base.strip("/")}/point/',
                        json={"query": description, "base64": screenshot_base64},
                    )
                    return response.json()
                except Exception as e:
                    raise Exception(
                        str(e)
                        + "\n\nIcon locating API not available, or we were unable to find the icon. Please try another method to find this icon."
                    )

    def find_text(self, text, screenshot=None):
        """
        Searches for specified text within a screenshot or the current screen if no screenshot is provided.
        """
        if screenshot == None:
            screenshot = self.screenshot(show=False)

        if not self.computer.offline:
            # Convert the screenshot to base64
            buffered = BytesIO()
            screenshot.save(buffered, format="PNG")
            screenshot_base64 = base64.b64encode(buffered.getvalue()).decode()

            try:
                response = requests.post(
                    f'{self.computer.api_base.strip("/")}/point/text/',
                    json={"query": text, "base64": screenshot_base64},
                )
                response = response.json()
                return response
            except:
                print("Attempting to find the text locally.")

        # We'll only get here if 1) self.computer.offline = True, or the API failed

        # Find the text in the screenshot
        centers = find_text_in_image(screenshot, text, self.computer.debug)

        return [
            {"coordinates": center, "text": "", "similarity": 1} for center in centers
        ]  # Have it deliver the text properly soon.

    def get_text_as_list_of_lists(self, screenshot=None):
        """
        Extracts and returns text from a screenshot or the current screen as a list of lists, each representing a line of text.
        """
        if screenshot == None:
            screenshot = self.screenshot(show=False, force_image=True)

        if not self.computer.offline:
            # Convert the screenshot to base64
            buffered = BytesIO()
            screenshot.save(buffered, format="PNG")
            screenshot_base64 = base64.b64encode(buffered.getvalue()).decode()

            try:
                response = requests.post(
                    f'{self.computer.api_base.strip("/")}/text/',
                    json={"base64": screenshot_base64},
                )
                response = response.json()
                return response
            except:
                print("Attempting to get the text locally.")

        # We'll only get here if 1) self.computer.offline = True, or the API failed

        try:
            return pytesseract_get_text(screenshot)
        except:
            raise Exception(
                "Failed to find text locally.\n\nTo find text in order to use the mouse, please make sure you've installed `pytesseract` along with the Tesseract executable (see this Stack Overflow answer for help installing Tesseract: https://stackoverflow.com/questions/50951955/pytesseract-tesseractnotfound-error-tesseract-is-not-installed-or-its-not-i)."
            )


def take_screenshot_to_pil(screen=0, combine_screens=True):
    # Get information about all screens
    monitors = screeninfo.get_monitors()
    if screen == -1:  # All screens
        # Take a screenshot of each screen and save them in a list
        screenshots = [
            pyautogui.screenshot(
                region=(monitor.x, monitor.y, monitor.width, monitor.height)
            )
            for monitor in monitors
        ]

        if combine_screens:
            # Combine all screenshots horizontally
            total_width = sum([img.width for img in screenshots])
            max_height = max([img.height for img in screenshots])

            # Create a new image with a size that can contain all screenshots
            new_img = Image.new("RGB", (total_width, max_height))

            # Paste each screenshot into the new image
            x_offset = 0
            for i, img in enumerate(screenshots):
                # Convert PIL Image to OpenCV Image (numpy array)
                img_cv = np.array(img)
                img_cv = cv2.cvtColor(img_cv, cv2.COLOR_RGB2BGR)

                # Convert new_img PIL Image to OpenCV Image (numpy array)
                new_img_cv = np.array(new_img)
                new_img_cv = cv2.cvtColor(new_img_cv, cv2.COLOR_RGB2BGR)

                # Paste each screenshot into the new image using OpenCV
                new_img_cv[
                    0 : img_cv.shape[0], x_offset : x_offset + img_cv.shape[1]
                ] = img_cv
                x_offset += img.width

                # Add monitor labels using OpenCV
                font = cv2.FONT_HERSHEY_SIMPLEX
                font_scale = 4
                font_color = (255, 255, 255)
                line_type = 2

                if i == 0:
                    text = "Primary Monitor"
                else:
                    text = f"Monitor {i}"

                # Calculate the font scale that will fit the text perfectly in the center of the monitor
                text_size = cv2.getTextSize(text, font, font_scale, line_type)[0]
                font_scale = min(img.width / text_size[0], img.height / text_size[1])

                # Recalculate the text size with the new font scale
                text_size = cv2.getTextSize(text, font, font_scale, line_type)[0]

                # Calculate the position to center the text
                text_x = x_offset - img.width // 2 - text_size[0] // 2
                text_y = max_height // 2 - text_size[1] // 2

                cv2.putText(
                    new_img_cv,
                    text,
                    (text_x, text_y),
                    font,
                    font_scale,
                    font_color,
                    line_type,
                )

                # Convert new_img from OpenCV Image back to PIL Image
                new_img_cv = cv2.cvtColor(new_img_cv, cv2.COLOR_BGR2RGB)
                new_img = Image.fromarray(new_img_cv)

            return new_img
        else:
            return screenshots
    elif screen > 0:
        # Take a screenshot of the selected screen
        return pyautogui.screenshot(
            region=(
                monitors[screen].x,
                monitors[screen].y,
                monitors[screen].width,
                monitors[screen].height,
            )
        )

    else:
        # Take a screenshot of the primary screen
        return pyautogui.screenshot(
            region=(
                monitors[screen].x,
                monitors[screen].y,
                monitors[screen].width,
                monitors[screen].height,
            )
        )


def get_displays():
    monitors = get_monitors()
    return monitors



import json

from .ai.ai import Ai
from .browser.browser import Browser
from .calendar.calendar import Calendar
from .clipboard.clipboard import Clipboard
from .contacts.contacts import Contacts
from .display.display import Display
from .docs.docs import Docs
from .files.files import Files
from .keyboard.keyboard import Keyboard
from .mail.mail import Mail
from .mouse.mouse import Mouse
from .os.os import Os
from .skills.skills import Skills
from .sms.sms import SMS
from .vision.vision import Vision

class Computer:
    def __init__(self):

        self.offline = False
        self.verbose = False
        self.debug = False

        self.mouse = Mouse(self)
        self.keyboard = Keyboard(self)
        self.display = Display(self)
        self.clipboard = Clipboard(self)
        self.mail = Mail(self)
        self.sms = SMS(self)
        self.calendar = Calendar(self)
        self.contacts = Contacts(self)
        self.browser = Browser(self)
        self.os = Os(self)
        self.vision = Vision(self)
        self.skills = Skills(self)
        self.docs = Docs(self)
        self.ai = Ai(self)
        self.files = Files(self)

        self.emit_images = True
        self.save_skills = True

        self.import_computer_api = False  # Defaults to false
        self._has_imported_computer_api = False  # Because we only want to do this once

        self.import_skills = False
        self._has_imported_skills = False

    @property
    def api_base(self):
        import os
        return os.environ.get('INTERPRETER_API_BASE')

    @property
    def api_key(self):
        import os
        return os.environ.get('INTERPRETER_API_KEY')

    def screenshot(self, *args, **kwargs):
        """
        Shortcut for computer.display.screenshot
        """
        return self.display.screenshot(*args, **kwargs)

    def view(self, *args, **kwargs):
        """
        Shortcut for computer.display.screenshot
        """
        return self.display.screenshot(*args, **kwargs)

    def to_dict(self):
        def json_serializable(obj):
            try:
                json.dumps(obj)
                return True
            except:
                return False

        return {k: v for k, v in self.__dict__.items() if json_serializable(v)}

    def load_dict(self, data_dict):
        for key, value in data_dict.items():
            if hasattr(self, key):
                setattr(self, key, value)



import glob
import inspect
import json
import os
import re
import subprocess
from pathlib import Path


from ..utils.lazy_import import lazy_import
from ..utils.recipient_utils import format_to_recipient

# Lazy import, imported when needed to speed up start time
aifs = lazy_import("aifs")
pyautogui = lazy_import("pyautogui")
pynput = lazy_import("pynput")

element = None
element_box = None
icon_dimensions = None


class Skills:
    def __init__(self, computer):
        self.computer = computer
        # Get the INTERPRETER_DIRECTORY environment variable
        oi_dir = os.environ.get('INTERPRETER_DIRECTORY')
        if not oi_dir:
            oi_dir = os.path.join(os.path.expanduser('~'), 'interpreter')
        self.path = str(Path(oi_dir) / "skills")
        self.new_skill = NewSkill(self)

    def list(self):
        return [
            file.replace(".py", "()")
            for file in os.listdir(self.path)
            if file.endswith(".py")
        ]

    def run(self, skill):
        print(
            "To run a skill, run its name as a function name (it is already imported)."
        )

    def search(self, query):
        """
        This just lists all for now.
        """
        return [
            file.replace(".py", "()")
            for file in os.listdir(self.path)
            if file.endswith(".py")
        ]

    def import_skills(self):
        previous_save_skills_setting = self.computer.save_skills

        self.computer.save_skills = False

        # Make sure it's not over 100mb
        total_size = 0
        for path, dirs, files in os.walk(self.path):
            for f in files:
                fp = os.path.join(path, f)
                total_size += os.path.getsize(fp)
        total_size = total_size / (1024 * 1024)  # convert bytes to megabytes
        if total_size > 100:
            raise Warning(
                f"Skills at path {self.path} can't exceed 100mb. Try deleting some."
            )

        code_to_run = ""
        for file in glob.glob(os.path.join(self.path, "*.py")):
            with open(file, "r") as f:
                code_to_run += f.read() + "\n"

        if self.computer.interpreter.debug:
            print("IMPORTING SKILLS:\n", code_to_run)

        output = self.computer.run("python", code_to_run)

        if "traceback" in str(output).lower():
            # Import them individually
            for file in glob.glob(os.path.join(self.path, "*.py")):
                with open(file, "r") as f:
                    code_to_run = f.read() + "\n"

                if self.computer.interpreter.debug:
                    print(self.path)
                    print("IMPORTING SKILL:\n", code_to_run)

                output = self.computer.run("python", code_to_run)

                if "traceback" in str(output).lower():
                    print(
                        f"Skill at {file} might be broken— it produces a traceback when run."
                    )

        self.computer.save_skills = previous_save_skills_setting


class NewSkill:
    def __init__(self, skills):
        self.path = ""
        self.skills = skills

    def create(self):
        self.steps = []
        self._name = "Untitled"
        print(
            """

INSTRUCTIONS
You are creating a new skill. Follow these steps exactly to get me to tell you its name:
1. Ask me what the name of this skill is.
2. After I explicitly tell you the name of the skill (I may tell you to proceed which is not the name— if I do say that, you probably need more information from me, so tell me that), after you get the proper name, execute `computer.skills.new_skill.name = "{INSERT THE SKILL NAME FROM QUESTION #1}"`.
        
        """.strip()
        )

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        print(
            """

Skill named. Now, follow these next INSTRUCTIONS exactly:

1. Ask me what the first step is.
2. When I reply, execute code to accomplish that step. Write comments explaining your reasoning before each line.
3. Ask me if you completed the step correctly.
    a. (!!!!!!!!!!!! >>>>>> THIS IS CRITICAL. DO NOT FORGET THIS.) IF you completed it correctly, run `computer.skills.new_skill.add_step(step, code)` where step is a generalized, natural language description of the step, and code is the code you ran to complete it.
    b. IF you did not complete it correctly, try to fix your code and ask me again.
4. If I say the skill is complete, or that that was the last step, run `computer.skills.new_skill.save()`.

YOU MUST FOLLOW THESE 4 INSTRUCTIONS **EXACTLY**. I WILL TIP YOU $200.

              """.strip()
        )

    def add_step(self, step, code):
        self.steps.append(step + "\n\n```python\n" + code + "\n```")
        print(
            """

Step added. Now, follow these next INSTRUCTIONS exactly:

1. Ask me what the next step is.
2. When I reply, execute code to accomplish that step.
3. Ask me if you completed the step correctly.
    a. (!!!!!!!!!!!! >>>>>> THIS IS CRITICAL. DO NOT FORGET THIS!!!!!!!!.) IF you completed it correctly, run `computer.skills.new_skill.add_step(step, code)` where step is a generalized, natural language description of the step, and code is the code you ran to complete it.
    b. IF you did not complete it correctly, try to fix your code and ask me again.
4. If I say the skill is complete, or that that was the last step, run `computer.skills.new_skill.save()`.

YOU MUST FOLLOW THESE 4 INSTRUCTIONS **EXACTLY**. I WILL TIP YOU $200.

        """.strip()
        )

    def save(self):
        normalized_name = re.sub("[^0-9a-zA-Z]+", "_", self.name.lower())

        skill_string = f'''
import json

def {normalized_name}(step=0):
    """
    Run this function to {normalized_name}. Pass in step=0 to see the first step, step=1 to see the next step, etc.
    """
    steps = {self.steps}

    print("")

    if step < len(steps):
        if isinstance(steps[step], str):
            print("To complete this task / run this skill, flexibly complete the following step, swapping out parts as necessary to fulfill the user's task. You will need to run the following code yourself, it hasn't run yet!")
            print("Step " + str(step + 1) + ": " + steps[step])
        else:
            computer.mouse.click(steps[step]["element"], icon_dimensions=steps[step]["icon_dimensions"]) # Instructed click
        if step + 1 < len(steps):
            print("After completing the above, I need you to run {normalized_name}(step=" + str(step + 1) + ") immediatly.")
        else:
            print("After executing the code, you have completed all the steps, the task/skill has been run!")
    else:
        print("The specified step number exceeds the available steps. Please run with a valid step number.")
'''.strip()

        skill_file_path = os.path.join(self.skills.path, f"{normalized_name}.py")

        if not os.path.exists(self.skills.path):
            os.makedirs(self.skills.path)

        with open(skill_file_path, "w") as file:
            file.write(skill_string)

        # Execute the code in skill_string to define the function
        exec(skill_string)

        # Verify that the file was written
        if os.path.exists(skill_file_path):
            print("SKILL SAVED:", self.name.upper())
            print(
                "Teaching session finished. Tell the user that the skill above has been saved. Great work!"
            )
        else:
            print(f"Error: Failed to write skill file to {skill_file_path}")



import base64
import contextlib
import io
import os
import tempfile

from PIL import Image

from ..utils.lazy_import import lazy_import
from ..utils.computer_vision import pytesseract_get_text

# transformers = lazy_import("transformers") # Doesn't work for some reason! We import it later.


class Vision:
    def __init__(self, computer):
        self.computer = computer
        self.model = None  # Will load upon first use
        self.tokenizer = None  # Will load upon first use
        self.easyocr = None

    def load(self, load_moondream=True, load_easyocr=True):
        # print("Loading vision models (Moondream, EasyOCR)...\n")

        with contextlib.redirect_stdout(
            open(os.devnull, "w")
        ), contextlib.redirect_stderr(open(os.devnull, "w")):
            if self.easyocr == None and load_easyocr:
                import easyocr

                self.easyocr = easyocr.Reader(
                    ["en"]
                )  # this needs to run only once to load the model into memory

            if self.model == None and load_moondream:
                import transformers  # Wait until we use it. Transformers can't be lazy loaded for some reason!

                os.environ["TOKENIZERS_PARALLELISM"] = "false"

                if self.computer.debug:
                    print(
                        "Open Interpreter will use Moondream (tiny vision model) to describe images to the language model. Set `interpreter.llm.vision_renderer = None` to disable this behavior."
                    )
                    print(
                        "Alternatively, you can use a vision-supporting LLM and set `interpreter.llm.supports_vision = True`."
                    )
                model_id = "vikhyatk/moondream2"
                revision = "2024-04-02"
                print("loading model")

                self.model = transformers.AutoModelForCausalLM.from_pretrained(
                    model_id, trust_remote_code=True, revision=revision
                )
                self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                    model_id, revision=revision
                )
                return True

    def ocr(
        self,
        base_64=None,
        path=None,
        lmc=None,
        pil_image=None,
    ):
        """
        Gets OCR of image.
        """

        if lmc:
            if "base64" in lmc["format"]:
                # # Extract the extension from the format, default to 'png' if not specified
                # if "." in lmc["format"]:
                #     extension = lmc["format"].split(".")[-1]
                # else:
                #     extension = "png"
                # Save the base64 content as a temporary file
                img_data = base64.b64decode(lmc["content"])
                with tempfile.NamedTemporaryFile(
                    delete=False, suffix=".png"
                ) as temp_file:
                    temp_file.write(img_data)
                    temp_file_path = temp_file.name

                # Set path to the path of the temporary file
                path = temp_file_path

            elif lmc["format"] == "path":
                # Convert to base64
                path = lmc["content"]
        elif base_64:
            # Save the base64 content as a temporary file
            img_data = base64.b64decode(base_64)
            with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as temp_file:
                temp_file.write(img_data)
                temp_file_path = temp_file.name

            # Set path to the path of the temporary file
            path = temp_file_path
        elif path:
            pass
        elif pil_image:
            with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as temp_file:
                pil_image.save(temp_file, format="PNG")
                temp_file_path = temp_file.name

            # Set path to the path of the temporary file
            path = temp_file_path

        try:
            if not self.easyocr:
                self.load(load_moondream=False)
            result = self.easyocr.readtext(path)
            text = " ".join([item[1] for item in result])
            return text.strip()
        except ImportError:
            print(
                "\nTo use local vision, run `pip install 'open-interpreter[local]'`.\n"
            )
            return ""

    def query(
        self,
        query="Describe this image. Also tell me what text is in the image, if any.",
        base_64=None,
        path=None,
        lmc=None,
        pil_image=None,
    ):
        """
        Uses Moondream to ask query of the image (which can be a base64, path, or lmc message)
        """

        if self.model == None and self.tokenizer == None:
            try:
                success = self.load(load_easyocr=False)
            except ImportError:
                print(
                    "\nTo use local vision, run `pip install 'open-interpreter[local]'`.\n"
                )
                return ""
            if not success:
                return ""

        if lmc:
            if "base64" in lmc["format"]:
                # # Extract the extension from the format, default to 'png' if not specified
                # if "." in lmc["format"]:
                #     extension = lmc["format"].split(".")[-1]
                # else:
                #     extension = "png"

                # Decode the base64 image
                img_data = base64.b64decode(lmc["content"])
                img = Image.open(io.BytesIO(img_data))

            elif lmc["format"] == "path":
                # Convert to base64
                image_path = lmc["content"]
                img = Image.open(image_path)
        elif base_64:
            img_data = base64.b64decode(base_64)
            img = Image.open(io.BytesIO(img_data))
        elif path:
            img = Image.open(path)
        elif pil_image:
            img = pil_image

        with contextlib.redirect_stdout(open(os.devnull, "w")):
            enc_image = self.model.encode_image(img)
            answer = self.model.answer_question(
                enc_image, query, self.tokenizer, max_length=400
            )

        return answer



import os
import platform
import re
import subprocess

from ..utils.run_applescript import run_applescript, run_applescript_capture


class Mail:
    def __init__(self, computer):
        self.computer = computer
        # In the future, we should allow someone to specify their own mail app
        self.mail_app = "Mail"

    def get(self, number=5, unread: bool = False):
        """
        Retrieves the last {number} emails from the inbox, optionally filtering for only unread emails.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        too_many_emails_msg = ""
        if number > 50:
            number = min(number, 50)
            too_many_emails_msg = (
                "This method is limited to 10 emails, returning the first 10: "
            )
        # This is set up to retry if the number of emails is less than the number requested, but only a max of three times
        retries = 0  # Initialize the retry counter
        while retries < 3:
            read_status_filter = "whose read status is false" if unread else ""
            script = f"""
            tell application "{self.mail_app}"
                set latest_messages to messages of inbox {read_status_filter}
                set email_data to {{}}
                repeat with i from 1 to {number}
                    set this_message to item i of latest_messages
                    set end of email_data to {{subject:subject of this_message, sender:sender of this_message, content:content of this_message}}
                end repeat
                return email_data
            end tell
            """
            stdout, stderr = run_applescript_capture(script)

            # if the error is due to not having enough emails, retry with the available emails.
            if "Can’t get item" in stderr:
                match = re.search(r"Can’t get item (\d+) of", stderr)
                if match:
                    available_emails = int(match.group(1)) - 1
                    if available_emails > 0:
                        number = available_emails
                        retries += 1
                        continue
                break
            elif stdout:
                if too_many_emails_msg:
                    return f"{too_many_emails_msg}\n\n{stdout}"
                else:
                    return stdout

    def send(self, to, subject, body, attachments=None):
        """
        Sends an email with the given parameters using the default mail app.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        # Strip newlines from the to field
        to = to.replace("\n", "")

        attachment_clause = ""
        delay_seconds = 5  # Default delay in seconds

        if attachments:
            formatted_attachments = [
                self.format_path_for_applescript(path) for path in attachments
            ]

            # Generate AppleScript to attach each file
            attachment_clause = "\n".join(
                f"make new attachment with properties {{file name:{path}}} at after the last paragraph of the content of new_message"
                for path in formatted_attachments
            )

            # Calculate the delay based on the size of the attachments
            delay_seconds = self.calculate_upload_delay(attachments)

            print(f"Uploading attachments. This should take ~{delay_seconds} seconds.")

        # In the future, we might consider allowing the llm to specify an email to send from
        script = f"""
        tell application "{self.mail_app}"
            set new_message to make new outgoing message with properties {{subject:"{subject}", content:"{body}"}} at end of outgoing messages
            tell new_message
                set visible to true
                make new to recipient at end of to recipients with properties {{address:"{to}"}}
                {attachment_clause}
            end tell
            {f'delay {delay_seconds}' if attachments else ''}
            send new_message
        end tell
        """
        try:
            run_applescript(script)
            return f"""Email sent to {to}"""
        except subprocess.CalledProcessError:
            return "Failed to send email"

    def unread_count(self):
        """
        Retrieves the count of unread emails in the inbox, limited to 50.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        script = f"""
            tell application "{self.mail_app}"
                set unreadMessages to (messages of inbox whose read status is false)
                if (count of unreadMessages) > 50 then
                    return 50
                else
                    return count of unreadMessages
                end if
            end tell
            """
        try:
            unreads = int(run_applescript(script))
            if unreads >= 50:
                return "50 or more"
            return unreads
        except subprocess.CalledProcessError as e:
            print(e)
            return 0

    # Estimate how long something will take to upload
    def calculate_upload_delay(self, attachments):
        try:
            total_size_mb = sum(
                os.path.getsize(os.path.expanduser(att)) for att in attachments
            ) / (1024 * 1024)
            # Assume 1 MBps upload speed, which is conservative on purpose
            upload_speed_mbps = 1
            estimated_time_seconds = total_size_mb / upload_speed_mbps
            return round(
                max(0.2, estimated_time_seconds + 1), 1
            )  # Add 1 second buffer, ensure a minimum delay of 1.2 seconds, rounded to one decimal place
        except:
            # Return a default delay of 5 seconds if an error occurs
            return 5

    def format_path_for_applescript(self, file_path):
        # Escape backslashes, quotes, and curly braces for AppleScript
        file_path = (
            file_path.replace("\\", "\\\\")
            .replace('"', '\\"')
            .replace("{", "\\{")
            .replace("}", "\\}")
        )
        # Convert to a POSIX path and quote for AppleScript
        posix_path = f'POSIX file "{file_path}"'
        return posix_path



import hashlib
import io
import os
import subprocess
from typing import List

import cv2
import nltk
import numpy as np
import torch
from PIL import Image, ImageDraw, ImageEnhance, ImageFont
from sentence_transformers import SentenceTransformer, util

from .....terminal_interface.utils.oi_dir import oi_dir
from ...utils.computer_vision import pytesseract_get_text_bounding_boxes

try:
    nltk.corpus.words.words()
except LookupError:
    nltk.download("words", quiet=True)
from nltk.corpus import words

# Create a set of English words
english_words = set(words.words())


def take_screenshot_to_pil(filename="temp_screenshot.png"):
    # Capture the screenshot and save it to a temporary file
    subprocess.run(["screencapture", "-x", filename], check=True)

    # Open the image file with PIL
    with open(filename, "rb") as f:
        image_data = f.read()
    image = Image.open(io.BytesIO(image_data))

    # Optionally, delete the temporary file if you don't need it after loading
    os.remove(filename)

    return image


from ...utils.computer_vision import find_text_in_image


def point(description, screenshot=None, debug=False, hashes=None):
    if description.startswith('"') and description.endswith('"'):
        return find_text_in_image(description.strip('"'), screenshot, debug)
    else:
        return find_icon(description, screenshot, debug, hashes)


def find_icon(description, screenshot=None, debug=False, hashes=None):
    if debug:
        print("STARTING")
    if screenshot == None:
        image_data = take_screenshot_to_pil()
    else:
        image_data = screenshot

    if hashes == None:
        hashes = {}

    image_width, image_height = image_data.size

    # Create a temporary file to save the image data
    #   with tempfile.NamedTemporaryFile(delete=False, suffix='.png') as temp_file:
    #     temp_file.write(base64.b64decode(request.base64))
    #     temp_image_path = temp_file.name
    #   print("yeah took", time.time()-thetime)

    icons_bounding_boxes = get_element_boxes(image_data, debug)

    if debug:
        print("GOT ICON BOUNDING BOXES")

    debug_path = os.path.join(os.path.expanduser("~"), "Desktop", "oi-debug")

    if debug:
        # Create a draw object
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw red rectangles around all blocks
        for block in icons_bounding_boxes:
            left, top, width, height = (
                block["x"],
                block["y"],
                block["width"],
                block["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="red")
        image_data_copy.save(
            os.path.join(debug_path, "before_filtering_out_extremes.png")
        )

    # Filter out extremes
    min_icon_width = int(os.getenv("OI_POINT_MIN_ICON_WIDTH", "10"))
    max_icon_width = int(os.getenv("OI_POINT_MAX_ICON_WIDTH", "500"))
    min_icon_height = int(os.getenv("OI_POINT_MIN_ICON_HEIGHT", "10"))
    max_icon_height = int(os.getenv("OI_POINT_MAX_ICON_HEIGHT", "500"))
    icons_bounding_boxes = [
        box
        for box in icons_bounding_boxes
        if min_icon_width <= box["width"] <= max_icon_width
        and min_icon_height <= box["height"] <= max_icon_height
    ]

    if debug:
        # Create a draw object
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw red rectangles around all blocks
        for block in icons_bounding_boxes:
            left, top, width, height = (
                block["x"],
                block["y"],
                block["width"],
                block["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="red")
        image_data_copy.save(
            os.path.join(debug_path, "after_filtering_out_extremes.png")
        )

    # Compute center_x and center_y for each box
    for box in icons_bounding_boxes:
        box["center_x"] = box["x"] + box["width"] / 2
        box["center_y"] = box["y"] + box["height"] / 2

    # # Filter out text

    if debug:
        print("GETTING TEXT")

    response = pytesseract_get_text_bounding_boxes(screenshot)

    if debug:
        print("GOT TEXT, processing it")

    if debug:
        # Create a draw object
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw red rectangles around all blocks
        for block in response:
            left, top, width, height = (
                block["left"],
                block["top"],
                block["width"],
                block["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="blue")

        # Save the image to the desktop
        if not os.path.exists(debug_path):
            os.makedirs(debug_path)
        image_data_copy.save(os.path.join(debug_path, "pytesseract_blocks_image.png"))

    blocks = [
        b for b in response if len(b["text"]) > 2
    ]  # icons are sometimes text, like "X"

    # Filter blocks so the text.lower() needs to be a real word in the English dictionary
    filtered_blocks = []
    for b in blocks:
        words = b["text"].lower().split()
        words = [
            "".join(e for e in word if e.isalnum()) for word in words
        ]  # remove punctuation
        if all(word in english_words for word in words):
            filtered_blocks.append(b)
    blocks = filtered_blocks

    if debug:
        # Create a draw object
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw red rectangles around all blocks
        for block in blocks:
            left, top, width, height = (
                block["left"],
                block["top"],
                block["width"],
                block["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="green")
        image_data_copy.save(
            os.path.join(debug_path, "pytesseract_filtered_blocks_image.png")
        )

    if debug:
        # Create a draw object
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw red rectangles around all blocks
        for block in blocks:
            left, top, width, height = (
                block["left"],
                block["top"],
                block["width"],
                block["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="green")
            # Draw the detected text in the rectangle in small font
            # Use PIL's built-in bitmap font
            font = ImageFont.load_default()
            draw.text(
                (block["left"], block["top"]), block["text"], fill="red", font=font
            )
        image_data_copy.save(
            os.path.join(debug_path, "pytesseract_filtered_blocks_image_with_text.png")
        )

    # Create an empty list to store the filtered boxes
    filtered_boxes = []

    # Filter out boxes that fall inside text
    for box in icons_bounding_boxes:
        if not any(
            text_box["left"] <= box["x"] <= text_box["left"] + text_box["width"]
            and text_box["top"] <= box["y"] <= text_box["top"] + text_box["height"]
            and text_box["left"]
            <= box["x"] + box["width"]
            <= text_box["left"] + text_box["width"]
            and text_box["top"]
            <= box["y"] + box["height"]
            <= text_box["top"] + text_box["height"]
            for text_box in blocks
        ):
            filtered_boxes.append(box)
        else:
            pass
            # print("Filtered out an icon because I think it is text.")

    icons_bounding_boxes = filtered_boxes

    if debug:
        # Create a copy of the image data
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw green rectangles around all filtered boxes
        for box in filtered_boxes:
            left, top, width, height = (
                box["x"],
                box["y"],
                box["width"],
                box["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="green")
        # Save the image with the drawn rectangles
        image_data_copy.save(
            os.path.join(debug_path, "pytesseract_filtered_boxes_image.png")
        )

    # Filter out boxes that intersect with text at all
    filtered_boxes = []
    for box in icons_bounding_boxes:
        if not any(
            max(text_box["left"], box["x"])
            < min(text_box["left"] + text_box["width"], box["x"] + box["width"])
            and max(text_box["top"], box["y"])
            < min(text_box["top"] + text_box["height"], box["y"] + box["height"])
            for text_box in blocks
        ):
            filtered_boxes.append(box)
    icons_bounding_boxes = filtered_boxes

    if debug:
        # Create a copy of the image data
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        # Draw green rectangles around all filtered boxes
        for box in icons_bounding_boxes:
            left, top, width, height = (
                box["x"],
                box["y"],
                box["width"],
                box["height"],
            )
            draw.rectangle([(left, top), (left + width, top + height)], outline="green")
        # Save the image with the drawn rectangles
        image_data_copy.save(
            os.path.join(debug_path, "debug_image_after_filtering_boxes.png")
        )

    # # (DISABLED)
    # # Filter to the most icon-like dimensions

    # # Desired dimensions
    # desired_width = 30
    # desired_height = 30

    # # Calculating the distance of each box's dimensions from the desired dimensions
    # for box in icons_bounding_boxes:
    #     width_diff = abs(box["width"] - desired_width)
    #     height_diff = abs(box["height"] - desired_height)
    #     # Sum of absolute differences as a simple measure of "closeness"
    #     box["distance"] = width_diff + height_diff

    # # Sorting the boxes based on their closeness to the desired dimensions
    # sorted_boxes = sorted(icons_bounding_boxes, key=lambda x: x["distance"])

    # # Selecting the top 150 closest boxes
    # icons_bounding_boxes = sorted_boxes  # DISABLED [:150]

    # Expand a little

    # Define the pixel expansion amount
    pixel_expand = int(os.getenv("OI_POINT_PIXEL_EXPAND", 7))

    # Expand each box by pixel_expand
    for box in icons_bounding_boxes:
        # Expand x, y by pixel_expand if they are greater than 0
        box["x"] = box["x"] - pixel_expand if box["x"] - pixel_expand >= 0 else box["x"]
        box["y"] = box["y"] - pixel_expand if box["y"] - pixel_expand >= 0 else box["y"]

        # Expand w, h by pixel_expand, but not beyond image_width and image_height
        box["width"] = (
            box["width"] + pixel_expand * 2
            if box["x"] + box["width"] + pixel_expand * 2 <= image_width
            else image_width - box["x"] - box["width"]
        )
        box["height"] = (
            box["height"] + pixel_expand * 2
            if box["y"] + box["height"] + pixel_expand * 2 <= image_height
            else image_height - box["y"] - box["height"]
        )

    # Save a debug image with a descriptive name for the step we just went through
    if debug:
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        for box in icons_bounding_boxes:
            left = box["x"]
            top = box["y"]
            width = box["width"]
            height = box["height"]
            draw.rectangle([(left, top), (left + width, top + height)], outline="red")
        image_data_copy.save(
            os.path.join(debug_path, "debug_image_after_expanding_boxes.png")
        )

    def combine_boxes(icons_bounding_boxes):
        while True:
            combined_boxes = []
            for box in icons_bounding_boxes:
                for i, combined_box in enumerate(combined_boxes):
                    if (
                        box["x"] < combined_box["x"] + combined_box["width"]
                        and box["x"] + box["width"] > combined_box["x"]
                        and box["y"] < combined_box["y"] + combined_box["height"]
                        and box["y"] + box["height"] > combined_box["y"]
                    ):
                        combined_box["x"] = min(box["x"], combined_box["x"])
                        combined_box["y"] = min(box["y"], combined_box["y"])
                        combined_box["width"] = (
                            max(
                                box["x"] + box["width"],
                                combined_box["x"] + combined_box["width"],
                            )
                            - combined_box["x"]
                        )
                        combined_box["height"] = (
                            max(
                                box["y"] + box["height"],
                                combined_box["y"] + combined_box["height"],
                            )
                            - combined_box["y"]
                        )
                        break
                else:
                    combined_boxes.append(box.copy())
            if len(combined_boxes) == len(icons_bounding_boxes):
                break
            else:
                icons_bounding_boxes = combined_boxes
        return combined_boxes

    if os.getenv("OI_POINT_OVERLAP", "True") == "True":
        icons_bounding_boxes = combine_boxes(icons_bounding_boxes)

    if debug:
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        for box in icons_bounding_boxes:
            x, y, w, h = box["x"], box["y"], box["width"], box["height"]
            draw.rectangle([(x, y), (x + w, y + h)], outline="blue")
        image_data_copy.save(
            os.path.join(debug_path, "debug_image_after_combining_boxes.png")
        )

    icons = []
    for box in icons_bounding_boxes:
        x, y, w, h = box["x"], box["y"], box["width"], box["height"]

        icon_image = image_data.crop((x, y, x + w, y + h))

        # icon_image.show()
        # input("Press Enter to finish looking at the image...")

        icon = {}
        icon["data"] = icon_image
        icon["x"] = x
        icon["y"] = y
        icon["width"] = w
        icon["height"] = h

        icon_image_hash = hashlib.sha256(icon_image.tobytes()).hexdigest()
        icon["hash"] = icon_image_hash

        # Calculate the relative central xy coordinates of the bounding box
        center_x = box["center_x"] / image_width  # Relative X coordinate
        center_y = box["center_y"] / image_height  # Relative Y coordinate
        icon["coordinate"] = (center_x, center_y)

        icons.append(icon)

    # Draw and show an image with the full screenshot and all the icons bounding boxes drawn on it in red
    if debug:
        image_data_copy = image_data.copy()
        draw = ImageDraw.Draw(image_data_copy)
        for icon in icons:
            x, y, w, h = icon["x"], icon["y"], icon["width"], icon["height"]
            draw.rectangle([(x, y), (x + w, y + h)], outline="red")
        desktop = os.path.join(os.path.join(os.path.expanduser("~")), "Desktop")
        image_data_copy.save(os.path.join(desktop, "point_vision.png"))

    if "icon" not in description.lower():
        description += " icon"

    if debug:
        print("FINALLY, SEARCHING")

    top_icons = image_search(description, icons, hashes, debug)

    if debug:
        print("DONE")

    coordinates = [t["coordinate"] for t in top_icons]

    # Return the top pick icon data
    return coordinates


# torch.set_num_threads(4)

fast_model = True

# First, we load the respective CLIP model
model = SentenceTransformer("clip-ViT-B-32")


import os

import timm

if fast_model == False:
    # Check if the model file exists
    if not os.path.isfile(model_path):
        # If not, create and save the model
        model = timm.create_model(
            "vit_base_patch16_siglip_224",
            pretrained=True,
            num_classes=0,
        )
        model = model.eval()
        torch.save(model.state_dict(), model_path)
    else:
        # If the model file exists, load the model from the saved state
        model = timm.create_model(
            "vit_base_patch16_siglip_256",
            pretrained=False,  # Don't load pretrained weights
            num_classes=0,
        )
        model.load_state_dict(torch.load(model_path))
        model = model.eval()

    # get model specific transforms (normalization, resize)
    data_config = timm.data.resolve_model_data_config(model)
    transforms = timm.data.create_transform(**data_config, is_training=False)

    def embed_images(images: List[Image.Image], model, transforms):
        # Stack images along the batch dimension
        image_batch = torch.stack([transforms(image) for image in images])
        # Get embeddings
        embeddings = model(image_batch)
        return embeddings

    # Usage:
    # images = [Image.open(io.BytesIO(image_bytes1)), Image.open(io.BytesIO(image_bytes2)), ...]
    # embeddings = embed_images(images, model, transforms)


if torch.cuda.is_available():
    device = torch.device("cuda")
elif torch.backends.mps.is_available():
    device = torch.device("mps")
else:
    device = torch.device("cpu")

# Move the model to the specified device
model = model.to(device)


def image_search(query, icons, hashes, debug):
    hashed_icons = [icon for icon in icons if icon["hash"] in hashes]
    unhashed_icons = [icon for icon in icons if icon["hash"] not in hashes]

    # Embed the unhashed icons
    if fast_model:
        query_and_unhashed_icons_embeds = model.encode(
            [query] + [icon["data"] for icon in unhashed_icons],
            batch_size=128,
            convert_to_tensor=True,
            show_progress_bar=debug,
        )
    else:
        query_and_unhashed_icons_embeds = embed_images(
            [query] + [icon["data"] for icon in unhashed_icons], model, transforms
        )

    query_embed = query_and_unhashed_icons_embeds[0]
    unhashed_icons_embeds = query_and_unhashed_icons_embeds[1:]

    # Store hashes for unhashed icons
    for icon, emb in zip(unhashed_icons, unhashed_icons_embeds):
        hashes[icon["hash"]] = emb

    # Move tensors to the specified device before concatenating
    unhashed_icons_embeds = unhashed_icons_embeds.to(device)

    # Include hashed icons in img_emb
    img_emb = torch.cat(
        [unhashed_icons_embeds]
        + [hashes[icon["hash"]].unsqueeze(0) for icon in hashed_icons]
    )

    # Perform semantic search
    hits = util.semantic_search(query_embed, img_emb)[0]

    # Filter hits with score over 90
    results = [hit for hit in hits if hit["score"] > 90]

    # Ensure top result is included
    if hits and (hits[0] not in results):
        results.insert(0, hits[0])

    # Convert results to original icon format
    return [icons[hit["corpus_id"]] for hit in results]


def get_element_boxes(image_data, debug):
    desktop_path = os.path.join(os.path.expanduser("~"), "Desktop")
    debug_path = os.path.join(desktop_path, "oi-debug")

    if debug:
        if not os.path.exists(debug_path):
            os.makedirs(debug_path)

    # Re-import the original image for contrast adjustment
    # original_image = cv2.imread(image_path)

    # Convert the image to a format that PIL can work with
    # pil_image = Image.fromarray(cv2.cvtColor(original_image, cv2.COLOR_BGR2RGB))

    pil_image = image_data

    # Convert to grayscale
    pil_image = pil_image.convert("L")

    def process_image(
        pil_image,
        contrast_level=1.8,
        debug=False,
        debug_path=None,
        adaptive_method=cv2.ADAPTIVE_THRESH_MEAN_C,
        threshold_type=cv2.THRESH_BINARY_INV,
        block_size=11,
        C=3,
    ):
        # Apply an extreme contrast filter
        enhancer = ImageEnhance.Contrast(pil_image)
        contrasted_image = enhancer.enhance(
            contrast_level
        )  # Significantly increase contrast

        # Create a string with all parameters
        parameters_string = f"contrast_level_{contrast_level}-adaptive_method_{adaptive_method}-threshold_type_{threshold_type}-block_size_{block_size}-C_{C}"

        if debug:
            print("TRYING:", parameters_string)
            contrasted_image_path = os.path.join(
                debug_path, f"contrasted_image_{parameters_string}.jpg"
            )
            contrasted_image.save(contrasted_image_path)
            print(f"DEBUG: Contrasted image saved to {contrasted_image_path}")

        # Convert the contrast-enhanced image to OpenCV format
        contrasted_image_cv = cv2.cvtColor(
            np.array(contrasted_image), cv2.COLOR_RGB2BGR
        )

        # Convert the contrast-enhanced image to grayscale
        gray_contrasted = cv2.cvtColor(contrasted_image_cv, cv2.COLOR_BGR2GRAY)
        if debug:
            image_path = os.path.join(
                debug_path, f"gray_contrasted_image_{parameters_string}.jpg"
            )
            cv2.imwrite(image_path, gray_contrasted)
            print("DEBUG: Grayscale contrasted image saved at:", image_path)

        # Apply adaptive thresholding to create a binary image where the GUI elements are isolated
        binary_contrasted = cv2.adaptiveThreshold(
            src=gray_contrasted,
            maxValue=255,
            adaptiveMethod=adaptive_method,
            thresholdType=threshold_type,
            blockSize=block_size,
            C=C,
        )

        if debug:
            binary_contrasted_image_path = os.path.join(
                debug_path, f"binary_contrasted_image_{parameters_string}.jpg"
            )
            cv2.imwrite(binary_contrasted_image_path, binary_contrasted)
            print(
                f"DEBUG: Binary contrasted image saved to {binary_contrasted_image_path}"
            )

        # Find contours from the binary image
        contours_contrasted, _ = cv2.findContours(
            binary_contrasted, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE
        )

        # Optionally, draw contours on the image for visualization
        contour_image = np.zeros_like(binary_contrasted)
        cv2.drawContours(contour_image, contours_contrasted, -1, (255, 255, 255), 1)

        if debug:
            contoured_contrasted_image_path = os.path.join(
                debug_path, f"contoured_contrasted_image_{parameters_string}.jpg"
            )
            cv2.imwrite(contoured_contrasted_image_path, contour_image)
            print(
                f"DEBUG: Contoured contrasted image saved at: {contoured_contrasted_image_path}"
            )

        return contours_contrasted

    if os.getenv("OI_POINT_PERMUTATE", "False") == "True":
        import random

        for _ in range(10):
            random_contrast = random.uniform(
                1, 40
            )  # Random contrast in range 0.5 to 1.5
            random_block_size = random.choice(
                range(1, 11, 2)
            )  # Random block size in range 1 to 10, but only odd numbers
            random_block_size = 11
            random_adaptive_method = random.choice(
                [cv2.ADAPTIVE_THRESH_MEAN_C, cv2.ADAPTIVE_THRESH_GAUSSIAN_C]
            )  # Random adaptive method
            random_threshold_type = random.choice(
                [cv2.THRESH_BINARY, cv2.THRESH_BINARY_INV]
            )  # Random threshold type
            random_C = random.randint(-10, 10)  # Random C in range 1 to 10
            contours_contrasted = process_image(
                pil_image,
                contrast_level=random_contrast,
                block_size=random_block_size,
                adaptive_method=random_adaptive_method,
                threshold_type=random_threshold_type,
                C=random_C,
                debug=debug,
                debug_path=debug_path,
            )

        print("Random Contrast: ", random_contrast)
        print("Random Block Size: ", random_block_size)
        print("Random Adaptive Method: ", random_adaptive_method)
        print("Random Threshold Type: ", random_threshold_type)
        print("Random C: ", random_C)
    else:
        contours_contrasted = process_image(
            pil_image, debug=debug, debug_path=debug_path
        )

    if debug:
        print("WE HERE")

    # Initialize an empty list to store the boxes
    boxes = []
    for contour in contours_contrasted:
        # Get the rectangle that bounds the contour
        x, y, w, h = cv2.boundingRect(contour)
        # Append the box as a dictionary to the list
        boxes.append({"x": x, "y": y, "width": w, "height": h})

    if debug:
        print("WE HHERE")

    if (
        False
    ):  # Disabled. I thought this would be faster but it's actually slower than just embedding all of them.
        # Remove any boxes whose edges cross over any contours
        filtered_boxes = []
        for box in boxes:
            crosses_contour = False
            for contour in contours_contrasted:
                if (
                    cv2.pointPolygonTest(contour, (box["x"], box["y"]), False) >= 0
                    or cv2.pointPolygonTest(
                        contour, (box["x"] + box["width"], box["y"]), False
                    )
                    >= 0
                    or cv2.pointPolygonTest(
                        contour, (box["x"], box["y"] + box["height"]), False
                    )
                    >= 0
                    or cv2.pointPolygonTest(
                        contour,
                        (box["x"] + box["width"], box["y"] + box["height"]),
                        False,
                    )
                    >= 0
                ):
                    crosses_contour = True
                    break
            if not crosses_contour:
                filtered_boxes.append(box)
        boxes = filtered_boxes

    if debug:
        print("WE HHHERE")

    return boxes



import threading
import time

import html2text
import requests
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from webdriver_manager.chrome import ChromeDriverManager


class Browser:
    def __init__(self, computer):
        self.computer = computer
        self._driver = None

    @property
    def driver(self):
        if self._driver is None:
            self.setup()
        return self._driver

    @driver.setter
    def driver(self, value):
        self._driver = value

    def search(self, query):
        """
        Searches the web for the specified query and returns the results.
        """
        response = requests.get(
            f'{self.computer.api_base.strip("/")}/browser/search',
            params={"query": query},
        )
        return response.json()["result"]

    def fast_search(self, query):
        """
        Searches the web for the specified query and returns the results.
        """

        # Start the request in a separate thread
        response_thread = threading.Thread(
            target=lambda: setattr(
                threading.current_thread(),
                "response",
                requests.get(
                    f'{self.computer.api_base.strip("/")}/browser/search',
                    params={"query": query},
                ),
            )
        )
        response_thread.start()

        # Perform the Google search
        self.search_google(query, delays=False)

        # Wait for the request to complete and get the result
        response_thread.join()
        response = response_thread.response

        return response.json()["result"]

    def setup(self):
        self.service = Service(ChromeDriverManager().install())
        self.options = webdriver.ChromeOptions()
        self._driver = webdriver.Chrome(service=self.service, options=self.options)

    def go_to_url(self, url):
        """Navigate to a URL"""
        self.driver.get(url)
        time.sleep(1)

    def search_google(self, query, delays=True):
        """Perform a Google search"""
        self.driver.get("https://www.perplexity.ai")
        # search_box = self.driver.find_element(By.NAME, 'q')
        # search_box.send_keys(query)
        # search_box.send_keys(Keys.RETURN)
        body = self.driver.find_element(By.TAG_NAME, "body")
        body.send_keys(Keys.COMMAND + "k")
        time.sleep(0.5)
        active_element = self.driver.switch_to.active_element
        active_element.send_keys(query)
        active_element.send_keys(Keys.RETURN)
        if delays:
            time.sleep(3)

    def analyze_page(self, intent):
        """Extract HTML, list interactive elements, and analyze with AI"""
        html_content = self.driver.page_source
        text_content = html2text.html2text(html_content)

        # text_content = text_content[:len(text_content)//2]

        elements = (
            self.driver.find_elements(By.TAG_NAME, "a")
            + self.driver.find_elements(By.TAG_NAME, "button")
            + self.driver.find_elements(By.TAG_NAME, "input")
            + self.driver.find_elements(By.TAG_NAME, "select")
        )

        elements_info = [
            {
                "id": idx,
                "text": elem.text,
                "attributes": elem.get_attribute("outerHTML"),
            }
            for idx, elem in enumerate(elements)
        ]

        ai_query = f"""
        Below is the content of the current webpage along with interactive elements. 
        Given the intent "{intent}", please extract useful information and provide sufficient details 
        about interactive elements, focusing especially on those pertinent to the provided intent.
        
        If the information requested by the intent "{intent}" is present on the page, simply return that.

        If not, return the top 10 most relevant interactive elements in a concise, actionable format, listing them on separate lines
        with their ID, a description, and their possible action.

        Do not hallucinate.

        Page Content:
        {text_content}
        
        Interactive Elements:
        {elements_info}
        """

        # response = self.computer.ai.chat(ai_query)

        # screenshot = self.driver.get_screenshot_as_base64()
        # old_model = self.computer.interpreter.llm.model
        # self.computer.interpreter.llm.model = "gpt-4o-mini"
        # response = self.computer.ai.chat(ai_query, base64=screenshot)
        # self.computer.interpreter.llm.model = old_model

        old_model = self.computer.interpreter.llm.model
        self.computer.interpreter.llm.model = "gpt-4o-mini"
        response = self.computer.ai.chat(ai_query)
        self.computer.interpreter.llm.model = old_model

        print(response)
        print(
            "Please now utilize this information or interact with the interactive elements provided to answer the user's query."
        )

    def quit(self):
        """Close the browser"""
        self.driver.quit()



import datetime
import platform
import subprocess

from ..utils.run_applescript import run_applescript, run_applescript_capture


makeDateFunction = """
on makeDate(yr, mon, day, hour, min, sec)
	set theDate to current date
	tell theDate
		set its year to yr
		set its month to mon
		set its day to day
		set its hours to hour
		set its minutes to min
		set its seconds to sec
	end tell
	return theDate
end makeDate
"""

class Calendar:
    def __init__(self, computer):
        self.computer = computer
        # In the future, we might consider a way to use a different calendar app. For now its Calendar
        self.calendar_app = "Calendar"

    def get_events(self, start_date=datetime.date.today(), end_date=None):
        """
        Fetches calendar events for the given date or date range.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        if not end_date:
            end_date = start_date
        # AppleScript command
        script = f"""
        {makeDateFunction}
        set theDate to makeDate({start_date.strftime("%Y, %m, %d, 0, 0, 0")})
        set endDate to makeDate({end_date.strftime("%Y, %m, %d, 23, 59, 59")})
        tell application "System Events"
            set calendarIsRunning to (name of processes) contains "{self.calendar_app}"
            if calendarIsRunning then
                tell application "{self.calendar_app}" to activate
            else
                tell application "{self.calendar_app}" to launch
                delay 1 -- Wait for the application to open
                tell application "{self.calendar_app}" to activate
            end if
        end tell

        set outputText to ""

        -- Access the Calendar app
        tell application "{self.calendar_app}"
            
            -- Initialize a list to hold summaries and dates of all events from all calendars
            set allEventsInfo to {{}}
            
            -- Loop through each calendar
            repeat with aCalendar in calendars
                
                -- Fetch events from this calendar that fall within the specified date range
                set theseEvents to (every event of aCalendar where its start date is greater than theDate and its start date is less than endDate)
                
                -- Loop through theseEvents to extract necessary details
                repeat with anEvent in theseEvents
                    -- Initialize variables to "None" to handle missing information gracefully
                    set attendeesString to "None"
                    set theNotes to "None"
                    set theLocation to "None"
                    
                    -- Try to get attendees, but fail gracefully
                    try
                        set attendeeNames to {{}}
                        repeat with anAttendee in attendees of anEvent
                            set end of attendeeNames to name of anAttendee
                        end repeat
                        if (count of attendeeNames) > 0 then
                            set attendeesString to my listToString(attendeeNames, ", ")
                        end if
                    on error
                        set attendeesString to "None"
                    end try
                    
                    -- Try to get notes, but fail gracefully
                    try
                        set theNotes to notes of anEvent
                        if theNotes is missing value then set theNotes to "None"
                    on error
                        set theNotes to "None"
                    end try
                    
                    -- Try to get location, but fail gracefully
                    try
                        set theLocation to location of anEvent
                        if theLocation is missing value then set theLocation to "None"
                    on error
                        set theLocation to "None"
                    end try
                    
                    -- Create a record with the detailed information of the event
                    set eventInfo to {{|summary|:summary of anEvent, |startDate|:start date of anEvent, |endDate|:end date of anEvent, |attendees|:attendeesString, notes:theNotes, |location|:theLocation}}
                    -- Append this record to the allEventsInfo list
                    set end of allEventsInfo to eventInfo
                end repeat
            end repeat
        end tell

        -- Check if any events were found and build the output text
        if (count of allEventsInfo) > 0 then
            repeat with anEventInfo in allEventsInfo
                -- Always include Event, Start Date, and End Date
                set eventOutput to "Event: " & (summary of anEventInfo) & " | Start Date: " & (|startDate| of anEventInfo) & " | End Date: " & (|endDate| of anEventInfo)
                
                -- Conditionally include other details if they are not "None"
                if (attendees of anEventInfo) is not "None" then
                    set eventOutput to eventOutput & " | Attendees: " & (attendees of anEventInfo)
                end if
                if (notes of anEventInfo) is not "None" then
                    set eventOutput to eventOutput & " | Notes: " & (notes of anEventInfo)
                end if
                if (location of anEventInfo) is not "None" then
                    set eventOutput to eventOutput & " | Location: " & (location of anEventInfo)
                end if
                
                -- Add the event's output to the overall outputText, followed by a newline for separation
                set outputText to outputText & eventOutput & "
        "
            end repeat
        else
            set outputText to "No events found for the specified date."
        end if

        -- Return the output text
        return outputText

        -- Helper subroutine to convert a list to a string
        on listToString(theList, delimiter)
            set AppleScript's text item delimiters to delimiter
            set theString to theList as string
            set AppleScript's text item delimiters to ""
            return theString
        end listToString

        """

        # Get outputs from AppleScript
        stdout, stderr = run_applescript_capture(script)
        if stderr:
            # If the error is due to not having access to the calendar app, return a helpful message
            if "Not authorized to send Apple events to Calendar" in stderr:
                return "Calendar access not authorized. Please allow access in System Preferences > Security & Privacy > Automation."
            else:
                return stderr

        return stdout

    def create_event(
        self,
        title: str,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        location: str = "",
        notes: str = "",
        calendar: str = None,
    ) -> str:
        """
        Creates a new calendar event in the default calendar with the given parameters using AppleScript.
        """
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        # Format datetime for AppleScript
        applescript_start_date = start_date.strftime("%B %d, %Y %I:%M:%S %p")
        applescript_end_date = end_date.strftime("%B %d, %Y %I:%M:%S %p")

        # If there is no calendar, lets use the first calendar applescript returns. This should probably be modified in the future
        if calendar is None:
            calendar = self.get_first_calendar()
            if calendar is None:
                return "Can't find a default calendar. Please try again and specify a calendar name."

        script = f"""
        {makeDateFunction}
        set startDate to makeDate({start_date.strftime("%Y, %m, %d, %H, %M, %S")})
        set endDate to makeDate({end_date.strftime("%Y, %m, %d, %H, %M, %S")})
        -- Open and activate calendar first
        tell application "System Events"
            set calendarIsRunning to (name of processes) contains "{self.calendar_app}"
            if calendarIsRunning then
                tell application "{self.calendar_app}" to activate
            else
                tell application "{self.calendar_app}" to launch
                delay 1 -- Wait for the application to open
                tell application "{self.calendar_app}" to activate
            end if
        end tell
        tell application "{self.calendar_app}"
            tell calendar "{calendar}"
                make new event at end with properties {{summary:"{title}", start date:startDate, end date:endDate, location:"{location}", description:"{notes}"}}
            end tell
            -- tell the Calendar app to refresh if it's running, so the new event shows up immediately
            tell application "{self.calendar_app}" to reload calendars
        end tell
        """

        try:
            run_applescript(script)
            return f"""Event created successfully in the "{calendar}" calendar."""
        except subprocess.CalledProcessError as e:
            return str(e)

    def delete_event(
        self, event_title: str, start_date: datetime.datetime, calendar: str = None
    ) -> str:
        if platform.system() != "Darwin":
            return "This method is only supported on MacOS"

        # The applescript requires a title and start date to get the right event
        if event_title is None or start_date is None:
            return "Event title and start date are required"

        # If there is no calendar, lets use the first calendar applescript returns. This should probably be modified in the future
        if calendar is None:
            calendar = self.get_first_calendar()
            if not calendar:
                return "Can't find a default calendar. Please try again and specify a calendar name."

        script = f"""
        {makeDateFunction}
        set eventStartDate to makeDate({start_date.strftime("%Y, %m, %d, %H, %M, %S")})
        -- Open and activate calendar first
        tell application "System Events"
            set calendarIsRunning to (name of processes) contains "{self.calendar_app}"
            if calendarIsRunning then
                tell application "{self.calendar_app}" to activate
            else
                tell application "{self.calendar_app}" to launch
                delay 1 -- Wait for the application to open
                tell application "{self.calendar_app}" to activate
            end if
        end tell
        tell application "{self.calendar_app}"
            -- Specify the name of the calendar where the event is located
            set myCalendar to calendar "{calendar}"
            
            -- Define the exact start date and name of the event to find and delete
            set eventSummary to "{event_title}"
            
            -- Find the event by start date and summary
            set theEvents to (every event of myCalendar where its start date is eventStartDate and its summary is eventSummary)
            
            -- Check if any events were found
            if (count of theEvents) is equal to 0 then
                return "No matching event found to delete."
            else
                -- If the event is found, delete it
                repeat with theEvent in theEvents
                    delete theEvent
                end repeat
                save
                return "Event deleted successfully."
            end if
        end tell
        """

        stderr, stdout = run_applescript_capture(script)
        if stdout:
            return stdout[0].strip()
        elif stderr:
            if "successfully" in stderr:
                return stderr

            return f"""Error deleting event: {stderr}"""
        else:
            return "Unknown error deleting event. Please check event title and date."

    def get_first_calendar(self) -> str:
        # Literally just gets the first calendar name of all the calendars on the system. AppleScript does not provide a way to get the "default" calendar
        script = f"""
            -- Open calendar first
            tell application "System Events"
                set calendarIsRunning to (name of processes) contains "{self.calendar_app}"
                if calendarIsRunning is false then
                    tell application "{self.calendar_app}" to launch
                    delay 1 -- Wait for the application to open
                end if
            end tell
            tell application "{self.calendar_app}"
            -- Get the name of the first calendar
                set firstCalendarName to name of first calendar
            end tell
            return firstCalendarName
            """
        stdout = run_applescript_capture(script)
        if stdout:
            return stdout[0].strip()
        else:
            return None



import difflib
import os
import re
import traceback

class Files:
    def __init__(self, computer):
        self.computer = computer

    def _read_file(self, path):
        with open(path, "r") as file:
            data = file.read()
        return data

    def read(self, path=None, extensions=None):
        """
        In the future, this should just read any filetype. Returns images in a way the LLM understands if it's, for example, a PDF. For now, should turn any file into a string.
        """

        if path is None or path == ".":
            path = os.getcwd()

        if extensions is None:
            extensions = (
                ".txt", ".md", ".py", ".js", ".html", ".css", ".json", ".yaml", ".yml",
                ".ini", ".cfg", ".conf", ".toml", ".xml", ".csv", ".tsv",
                ".rst", ".adoc", ".tex", ".sh", ".bash", ".zsh", ".fish",
                ".sql", ".php", ".rb", ".pl", ".pm", ".go", ".java", ".kt", ".scala",
                ".c", ".cpp", ".h", ".hpp", ".cs", ".ts", ".swift", ".m", ".mm",
                ".r", ".lua", ".tcl", ".f", ".f90", ".hs", ".erl", ".ex", ".exs",
                ".clj", ".groovy", ".ps1", ".vbs", ".bat", ".cmd"
            )

        if os.path.isfile(path):
            return self._read_file(path)

        elif os.path.isdir(path):
            file_list = []
            files_that_dont_match_extension = []
            for root, _, files in os.walk(path):
                for file in files:
                    file_path = os.path.join(root, file)
                    if not file.endswith(extensions):
                        files_that_dont_match_extension.append(file_path)
                        print(f"Skipping file: {file_path} (does not match specified extensions)")
                        print("To include this file type, pass in a tuple of extensions to the 'extensions' parameter, e.g., extensions=('.py', '.txt', '.md')")
                        continue
                    try:
                        file_text = self._read_file(file_path)
                    except Exception as e:
                        print(
                            f"Failed to read file at path: {file_path}. Reason: {str(e)}"
                        )
                        continue
                    file_list.append({"path": file_path, "text": file_text})
            if file_list == []:
                raise Exception(f"No files with the extensions: {extensions} found in the path: {path}. Instead, these files were found there: {files_that_dont_match_extension}")
            return file_list  # "\n".join([f["path"] + "\n" + f["text"] for f in file_list])
        else:
            raise ValueError(f"Path '{path}' is neither a file nor a directory.")

    def search(self, query, path=None, height=3):
        """
        Search the filesystem for the given query.
        """
        if path is None or path == ".":
            path = os.getcwd()

        if os.path.isfile(path):
            try:
                with open(path, 'r', encoding='utf-8') as file:
                    lines = file.readlines()
            except UnicodeDecodeError:
                # If UTF-8 fails, try with a more permissive encoding
                with open(path, 'r', encoding='latin-1') as file:
                    lines = file.readlines()
            except Exception as e:
                raise Exception(f"Could not read path {path}: {str(e)}")
            matches = []
            for i, line in enumerate(lines):
                if query in line:
                    start = max(0, i - height)
                    end = min(len(lines), i + height + 1)
                    context = ''.join(lines[start:end])
                    if len(context) > 300:
                        index = context.find(query)
                        start = max(0, index - 150)
                        end = min(len(context), index + 150)
                        context = context[start:end]
                        if start > 0:
                            context = '...' + context
                        if end < len(context):
                            context = context + '...'
                    matches.append({
                        'path': path,
                        'line_number': i + 1,
                        'line': line.strip(),
                        'context': context
                    })
            return matches
        elif os.path.isdir(path):
            matches = []
            for root, _, files in os.walk(path):
                for file in files:
                    file_path = os.path.join(root, file)
                    matches.extend(self.search(query, file_path, height))
            if not matches:
                close_matches = get_close_matches_in_text(query, path)
                if close_matches:
                    raise ValueError(f"No exact matches found. Did you mean one of these? {close_matches}")
                else:
                    # print("Using flexible AI search, which might be slow...")
                    search_for = query
                    return self.files.query(query=f"Can you find all instances of something very close to this text:\n\n'''{search_for}'''\n\nWhere is that in the files? Be SPECIFIC as to the exact line(s) where it appears, and in what file(s).", path=path)
            else:
                return matches
        else:
            raise ValueError(f"Path '{path}' is neither a file nor a directory.")

    def query(self, query, path=None, extensions=None):
        text = self.read(path, extensions)
        answer = self.computer.ai.query(text, query)
        return answer

    def edit(self, path, original_text, replacement_text, force=False):
        """
        Edits a file on the filesystem, replacing the original text with the replacement text.
        """
        with open(path, "r") as file:
            filedata = file.read()

        if original_text not in filedata:
            matches = get_close_matches_in_text(original_text, filedata)
            if matches:
                suggestions = ", ".join(matches)
                raise ValueError(
                    f"Original text not found. Did you mean one of these? {suggestions}"
                )

        if force == False:
            
            for attempt in range(10):
                # Perform the replacement
                new_filedata = filedata.replace(original_text, replacement_text)

                syntax_errors = check_syntax(new_filedata)

                # Get context around the edit
                old_lines = filedata.splitlines()
                new_lines = new_filedata.splitlines()

                # Find the start and end of the edit
                edit_start = next(i for i, (old, new) in enumerate(zip(old_lines, new_lines)) if old != new)
                edit_end = next(i for i, (old, new) in enumerate(zip(old_lines[::-1], new_lines[::-1])) if old != new)
                edit_end = len(old_lines) - edit_end

                # Get 50 lines of context on either side
                context_start = max(0, edit_start - 50)
                context_end = min(len(new_lines), edit_end + 50)

                # Generate the diff
                diff = []
                diff.append("...")
                diff.append("")

                # Add 50 lines of context before the edit
                for i in range(context_start, edit_start):
                    diff.append(new_lines[i])

                diff.append("--- OLD CODE START ---")
                # Add all the old lines
                for i in range(edit_start, edit_end):
                    if i < len(old_lines):
                        diff.append(old_lines[i])
                diff.append("--- OLD CODE END ---")
                diff.append("--- NEW CODE START ---")
                # Add all the new lines
                for i in range(edit_start, edit_end):
                    if i < len(new_lines):
                        diff.append(new_lines[i])
                diff.append("--- NEW CODE END ---")

                # Add 50 lines of context after the edit
                for i in range(edit_end, context_end):
                    diff.append(new_lines[i])

                context = "\n".join(diff)

                # Prepare LLM query
                editor_system_message = "You are an expert Python programmer. Please review the following code edit and focus solely on ensuring the code is placed correctly within the file. The content of the code is correct, but its placement or indentation might be unintentional."
                editor_user_message = f"Please review this code edit, focusing only on whether its placement seems intentional or unintentional. Here's the context around the edit, including 50 lines before and after:\n\n{context}\n\n"

                if syntax_errors:
                    editor_user_message += f"This edit resulted in syntax errors. Here are the errors:\n\n{syntax_errors}\n\n"

                editor_user_message += "If the edit's placement appears to be intentional and correct, simply respond with 'Approved'. If you believe the placement is unintentional or incorrect, please provide suggestions to correct it wrapped in <find>old code</find> and <replace>new code</replace> tags. Focus only on placement and indentation, not on the content or style of the code."

                print("---EDIT QUERY---")
                print(editor_user_message)
                print("---END OF EDIT QUERY---")

                messages = [
                    {
                        "role": "system",
                        "type": "message",
                        "content": editor_system_message,
                    },
                    {"role": "user", "type": "message", "content": editor_user_message},
                ]
                
                print("\n--- APPLYING EDIT ---\n")
                report = self.computer.ai.chat(messages=messages, display=True)
                print("\n--- DONE APPLYING EDIT ---\n")

                # Check if the LLM approved the edit
                if "<find>" not in report and "<replace>" not in report:
                    break  # Exit the loop if the LLM approved the edit

                # Extract find and replace from the LLM response
                find_pattern = re.compile(r'<find>(.*?)</find>', re.DOTALL)
                replace_pattern = re.compile(r'<replace>(.*?)</replace>', re.DOTALL)
                
                find_match = find_pattern.search(report)
                replace_match = replace_pattern.search(report)
                
                if find_match and replace_match:
                    original_text = find_match.group(1)
                    replacement_text = replace_match.group(1)
                else:
                    print("LLM did not provide a valid fix or approval!")
                    continue
            else:
                raise ValueError(f"AI seemed to never approved this edit! Last AI message: {report}\n\nIf this seems to be an error, pass in force=True to computer.files.edit.")

        elif force == True:
            # Perform the replacement
            new_filedata = filedata.replace(original_text, replacement_text)

        # Apply the final approved edit
        with open(path, "w") as file:
            file.write(new_filedata)

    def map(self, path=None, extensions=None, codebase=None):
        if extensions == None:
            print("Defaulting extensions to ('.py'), pass in a tuple with other extensions to see more, but this should be sufficient.")
            extensions = ('.py')

        if path is None or path == ".":
            path = os.getcwd()
        
        if not os.path.isdir(path):
            raise ValueError(f"The provided path '{path}' is not a directory.")

        if codebase == None:
            codebase = is_codebase(path)

        def generate_tree(directory, prefix="", extensions=None):
            tree = ""
            contents = sorted(os.listdir(directory))
            for i, item in enumerate(contents):
                path_item = os.path.join(directory, item)
                if os.path.isdir(path_item):
                    sub_tree = generate_tree(path_item, prefix + "│   ", extensions)
                    if (
                        sub_tree.strip()
                    ):  # Only add the folder if it contains files with the specified extensions
                        tree += f"{prefix}├── {item}/\n"
                        tree += sub_tree
                else:
                    if extensions is None or item.endswith(extensions):
                        tree += f"{prefix}├── {item}\n"
                        if codebase and (
                            extensions is None or item.endswith(extensions)
                        ):
                            tree += generate_codebase_info(path_item, prefix + "│   ")
            return tree

        def generate_codebase_info(file_path, prefix=""):
            codebase_info = ""
            try:
                with open(file_path, "r", errors='ignore') as file:
                    lines = file.readlines()
                    for line in lines:
                        stripped_line = line.strip()
                        if stripped_line.startswith("class ") or stripped_line.startswith(
                            "def "
                        ):
                            codebase_info += f"{prefix}│   {stripped_line}\n"
            except UnicodeDecodeError:
                codebase_info += f"{prefix}│   [Error: Unable to read file]\n"
            except Exception as e:
                codebase_info += f"{prefix}│   [Error: {str(e)}]\n"
            return codebase_info

        return generate_tree(path, extensions=extensions)

def get_close_matches_in_text(original_text, filedata, n=3):
    """
    Returns the closest matches to the original text in the content of the file.
    """
    words = filedata.split()
    original_words = original_text.split()
    len_original = len(original_words)

    matches = []
    for i in range(len(words) - len_original + 1):
        phrase = " ".join(words[i : i + len_original])
        similarity = difflib.SequenceMatcher(None, original_text, phrase).ratio()
        matches.append((similarity, phrase))

    matches.sort(reverse=True)
    return [match[1] for match in matches[:n]]

def is_codebase(path):
    """
    Determine if the given path is part of a codebase (i.e., contains a .git directory).
    
    Args:
    path (str): The path to check. Can be a file or directory.
    
    Returns:
    bool: True if the path is part of a codebase, False otherwise.
    """
    # If it's a file, return False immediately
    if os.path.isfile(path):
        return False
    
    # Traverse up the directory tree
    while path != os.path.dirname(path):  # Stop at the root directory
        if os.path.exists(os.path.join(path, ".git")):
            return True
        path = os.path.dirname(path)
    
    return False

def check_syntax(code_string):
    try:
        compile(code_string, '<string>', 'exec')
        return None
    except SyntaxError:
        return traceback.format_exc()
